#ifndef MSP_DATAFILE_COLLECTION_H_
#define MSP_DATAFILE_COLLECTION_H_

#include <msp/core/maputils.h>
#include <msp/core/meta.h>
#include <msp/core/refptr.h>
#include "loader.h"

namespace Msp {
namespace DataFile {

/**
Helper struct to determine whether a Loader has a Collection typedef.
*/
template<typename T>
struct NeedsCollection
{
	struct Yes { char c[2]; };
	struct No { char c; };

	template<typename U>
	static Yes f(typename U::Collection *);
	template<typename U>
	static No f(...);

	enum { value = (sizeof(f<T>(0))==sizeof(Yes)) };
};

class CollectionItemTypeBase;

template<typename T>
class CollectionItemType;

/**
A collection of objects that can be loaded from a datafile.  Each object is
identified by a name, which must be unique across the entire collection.
*/
class Collection
{
public:
	/**
	Loads objects into a Collection.
	*/
	class Loader: public DataFile::Loader
	{
		template<typename T> friend class CollectionItemType;

	private:
		template<typename T, typename S, bool = NeedsCollection<typename T::Loader>::value>
		struct Add;

		Collection &coll;

	public:
		Loader(Collection &);
		Collection &get_object() const { return coll; }
	private:
		template<typename T, typename S, typename C>
		void coll_item(const std::string &n)
		{
			RefPtr<T> it = new T;
			load_sub(*it, dynamic_cast<C &>(coll));
			coll.add<S>(n, it.get());
			it.release();
		}

		template<typename T, typename S>
		void item(const std::string &n)
		{
			RefPtr<T> it = new T;
			load_sub(*it);
			coll.add<S>(n, it.get());
			it.release();
		}
	};

private:
	typedef std::map<std::string, Variant> ItemMap;
	typedef std::list<CollectionItemTypeBase *> TypeList;

	TypeList types;
	ItemMap items;

	Collection(const Collection &);
	Collection &operator=(const Collection &);
public:
	Collection() { }
	virtual ~Collection();

	/**
	Adds an object into the collection.  If a name collision occurs, an
	exception is thrown.  The collection takes ownership of the object.
	*/
	template<typename T>
	void add(const std::string &name, T *item)
	{
		if(!item)
			throw std::invalid_argument("Collection::add(item)");

		RefPtr<typename RemoveConst<T>::Type> ptr(item);
		try
		{
			insert_unique(items, name, ptr);
		}
		catch(...)
		{
			// Avoid deleting the object
			ptr.release();
			throw;
		}
	}

	/**
	Gets an object of a specific type from the collection.
	*/
	template<typename T>
	T &get(const std::string &name) const
	{
		typedef typename RemoveConst<T>::Type NCT;

		return *get_item(items, name).value<RefPtr<NCT> >();
	}

	/**
	Gets an object of a specific type from the collection.  If the name is not
	found in the collection and there is a creator for the item type, it is
	invoked.
	*/
	template<typename T>
	T &get(const std::string &);

	/**
	Returns a list of the names of objects of a specific type in the collection.
	*/
	template<typename T>
	std::list<std::string> get_names() const
	{
		std::list<std::string> result;
		for(ItemMap::const_iterator i=items.begin(); i!=items.end(); ++i)
			if(i->second.check_type<RefPtr<typename RemoveConst<T>::Type> >())
				result.push_back(i->first);
		return result;
	}

	/**
	Returns a list of objects of a specific type in the collection.
	*/
	template<typename T>
	std::list<T *> get_list() const
	{
		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

		std::list<T *> result;
		for(ItemMap::const_iterator i=items.begin(); i!=items.end(); ++i)
			if(i->second.check_type<RPNCT>())
				result.push_back(i->second.value<RPNCT>().get());
		return result;
	}

	/**
	Checks whether a name exists in the collection.  Does not care about the
	type of the object.
	*/
	bool contains(const std::string &n) const;

	/**
	Returns the name of an item in the collection.
	*/
	template<typename T>
	const std::string &get_name(T *d) const
	{
		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

		for(ItemMap::const_iterator i=items.begin(); i!=items.end(); ++i)
			if(i->second.check_type<RPNCT>())
				if(i->second.value<RPNCT>().get()==d)
					return i->first;
	
		// XXX Need better exception class
		throw std::runtime_error("Item not found in collection");
	}

protected:
	template<typename T>
	CollectionItemType<T> &add_type();
};


template<typename T, typename S>
struct Collection::Loader::Add<T, S, false>
{
	static void add(Loader &loader, const std::string &kwd)
	{ loader.add(kwd, &Loader::item<T, S>); }
};

template<typename T, typename S>
struct Collection::Loader::Add<T, S, true>
{
	static void add(Loader &loader, const std::string &kwd)
	{ loader.add(kwd, &Loader::coll_item<T, S, typename T::Loader::Collection>); }
};


class CollectionItemTypeBase
{
protected:
	class TagBase
	{
	protected:
		TagBase() { }
	public:
		virtual ~TagBase() { }
	};

	template<typename T>
	class Tag: public TagBase
	{
	public:
		virtual ~Tag() { }
	};

	std::string kwd;
	TagBase *tag;

	CollectionItemTypeBase(): tag(0) { }
public:
	virtual ~CollectionItemTypeBase()
	{ delete tag; }

	virtual void add_to_loader(Collection::Loader &) const = 0;
	virtual bool can_create() const = 0;
	virtual void create_item(Collection &, const std::string &) const = 0;

	template<typename T>
	bool check_type() const
	{ return dynamic_cast<Tag<T> *>(tag); }
};


template<typename T>
class CollectionItemType: public CollectionItemTypeBase
{
private:
	class CreatorBase
	{
	protected:
		CreatorBase() { }
	public:
		virtual ~CreatorBase() { }

		virtual T *create(Collection &, const std::string &) const = 0;
	};

	template<typename C>
	class Creator: public CreatorBase
	{
	public:
		typedef T *(C::*FuncPtr)(const std::string &);

	private:
		FuncPtr func;

	public:
		Creator(FuncPtr f): func(f) { }

		virtual T *create(Collection &coll, const std::string &name) const
		{ return (static_cast<C &>(coll).*func)(name); }
	};

	class StoreBase
	{
	protected:
		StoreBase() { }
	public:
		virtual ~StoreBase() { }

		virtual void store(Collection &, const std::string &, T *) = 0;

		virtual void add_to_loader(Collection::Loader &, const std::string &) = 0;
	};

	template<typename S>
	class Store: public StoreBase
	{
	public:
		virtual void store(Collection &coll, const std::string &name, T *obj)
		{ coll.add(name, static_cast<S *>(obj)); }

		virtual void add_to_loader(Collection::Loader &loader, const std::string &kwd)
		{ Collection::Loader::Add<T, S>::add(loader, kwd); }
	};

	CreatorBase *creat;
	StoreBase *store;

public:
	CollectionItemType():
		creat(0), store(new Store<T>)
	{ tag = new Tag<T>; }

	~CollectionItemType()
	{
		delete creat;
		delete store;
	}

	CollectionItemType &keyword(const std::string &k)
	{
		kwd = k;
		return *this;
	}

	template<typename C>
	CollectionItemType &creator(T *(C::*func)(const std::string &))
	{
		delete creat;
		creat = new Creator<C>(func);
		return *this;
	}

	template<typename S>
	CollectionItemType &store_as()
	{
		delete tag;
		tag = new Tag<S>;
		delete store;
		store = new Store<S>;
		return *this;
	}

	virtual void add_to_loader(Collection::Loader &loader) const
	{ store->add_to_loader(loader, kwd); }

	virtual bool can_create() const
	{ return creat!=0; }

	virtual void create_item(Collection &coll, const std::string &name) const
	{
		if(!creat)
			throw std::runtime_error("no creator");
		T *obj = creat->create(coll, name);
		if(obj)
			store->store(coll, name, obj);
	}
};


template<typename T>
T &Collection::get(const std::string &name)
{
	typedef typename RemoveConst<T>::Type NCT;

	if(!items.count(name))
	{
		for(TypeList::iterator i=types.begin(); i!=types.end(); ++i)
			if((*i)->can_create() && (*i)->check_type<NCT>())
				(*i)->create_item(*this, name);
	}

	return *get_item(items, name).value<RefPtr<NCT> >();
}

template<typename T>
CollectionItemType<T> &Collection::add_type()
{
	CollectionItemType<T> *type = new CollectionItemType<T>;
	types.push_back(type);
	return *type;
}

} // namespace DataFile
} // namespace Msp

#endif
